# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.context."""

from vizier._src.pyvizier.shared import context
from absl.testing import absltest


class ContextTest(absltest.TestCase):

  def testDefaultsNotShared(self):
    """Make sure default parameters are not shared between instances."""
    context1 = context.Context()
    context2 = context.Context()
    context1.parameters['x1'] = context.ParameterValue(5)
    self.assertEmpty(context2.parameters)


if __name__ == '__main__':
  absltest.main()
